/*
 * Tapioca XMPP Plugin
 * Copyright (c) 2005 INdT.
 * @author Andre Moreira Magalhaes <andre.magalhaes@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif 

#include "xmpp_connection.h"

#include "xmpp_audio_session.h"
#include "xmpp_presence.h"
#include "xmpp_privacy.h"
#include "xmpp_contact_list.h"
#include "xmpp_text_session.h"

#include "xmpp_call_c.h"
#include "xmpp_chat_c.h"
#include "xmpp_client_c.h"
#include "xmpp_ssl_c.h"
#include "xmpp_thread_c.h"

#include <tapioca/base/tapioca_enums.h>
#include <tapioca/base/tapioca_errors.h>
#include <tapioca/base/tapioca_utils.h>

#include <stdlib.h>
#include <string.h>

struct _XmppConnectionPrivate {
    xmpp_client_t client;
    XmppContactList *contact_list;
    XmppPresence *presence;
    XmppPrivacy *privacy;
    gint keepalive_event_id;
    gboolean disposed;
};

static void            xmpp_connection_dispose        (GObject *object);

static gboolean        xmpp_connection_connect        (TapiocaConnection *tc, 
                                                       const gchar *uri, 
                                                       const GHashTable *parameters, 
                                                       GError **error);
static void            xmpp_connection_disconnect     (TapiocaConnection *tc);
static TapiocaSession *xmpp_connection_create_session (TapiocaConnection *tc, 
                                                       guint type,
                                                       GError **error);
static TapiocaPresence *
                       xmpp_connection_get_presence   (TapiocaConnection *tc);

static TapiocaContactList *
                       xmpp_connection_get_contact_list   (TapiocaConnection *tc);

static TapiocaPrivacy *
                       xmpp_connection_get_privacy   (TapiocaConnection *tc);

static void            state_change_cb                (xmpp_client_t client,
                                                       gpointer data, 
                                                       gint state,
                                                       gint error);
static void            incoming_call_cb               (xmpp_client_t client, 
                                                       gpointer data, 
                                                       const gchar *from,
                                                       xmpp_call_t call);
static void            incoming_chat_cb               (xmpp_client_t client, 
                                                       gpointer data, 
                                                       const gchar *from,
                                                       xmpp_chat_t chat);

static gboolean        send_keepalive                 (xmpp_client_t client);
/* static gboolean        loop_thread                    (gpointer data); */

static xmpp_thread_t thread = NULL;

G_DEFINE_TYPE(XmppConnection, xmpp_connection, TAPIOCA_TYPE_CONNECTION)

static void 
xmpp_connection_class_init (XmppConnectionClass *klass)
{
    GObjectClass *gobject_class;
    TapiocaConnectionClass *tapioca_connection_class;

    gobject_class = (GObjectClass *) klass;
    tapioca_connection_class = (TapiocaConnectionClass*) klass;
    xmpp_connection_parent_class = g_type_class_peek_parent (klass);
    
    g_type_class_add_private (klass, sizeof (XmppConnectionPrivate));

    gobject_class->dispose = xmpp_connection_dispose;
    
    tapioca_connection_class->connect = xmpp_connection_connect;
    tapioca_connection_class->disconnect = xmpp_connection_disconnect;
    tapioca_connection_class->create_session = xmpp_connection_create_session;
    tapioca_connection_class->get_presence = xmpp_connection_get_presence;
    tapioca_connection_class->get_contact_list = xmpp_connection_get_contact_list;
    tapioca_connection_class->get_privacy = xmpp_connection_get_privacy;
}

static void 
xmpp_connection_init (XmppConnection *xc)
{
    xc->priv = XMPP_CONNECTION_GET_PRIVATE (xc);
    xc->priv->client = NULL;
    xc->priv->contact_list = xmpp_contact_list_new ();
    xc->priv->presence = xmpp_presence_new ();
    xc->priv->privacy = xmpp_privacy_new ();
    xc->priv->keepalive_event_id = -1;
    xc->priv->disposed = FALSE;
}

static void
xmpp_connection_dispose (GObject *object)
{
    XmppConnection *xc = XMPP_CONNECTION (object);

    if (xc->priv->disposed) {
        /* If dispose did already run, return. */
        return;
    }

    /* Make sure dispose does not run twice. */
    xc->priv->disposed = TRUE;
   
    G_OBJECT_CLASS (xmpp_connection_parent_class)->dispose (object);
}

static gboolean 
xmpp_connection_connect (TapiocaConnection *tc,
                         const gchar *uri,
                         const GHashTable *parameters,
                         GError **error)
{
    XmppConnection *xc = XMPP_CONNECTION (tc);
    gchar *username, *offset, *host;
    const gchar *user, *resource, *pass, *server, *proxy = NULL, *proxy_user = NULL, *proxy_pass = NULL;
    guint server_port, proxy_type = TAPIOCA_PROXY_TYPE_NONE, proxy_port = 0;
    gint use_encryption;
   
    username = g_strdup ((gchar *) g_hash_table_lookup ((GHashTable *) parameters, "username"));
    if (!username) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, "Invadid parameters");
        return FALSE;
    }

    offset = strchr (username, '@');
    if (!offset) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, "Invalid username");
        return FALSE;
    }
    
    host = g_strdup (offset + 1);
    user = username;
    *offset = '\0';

    resource = (gchar *) g_hash_table_lookup ((GHashTable *) parameters, "resource");
    pass = (gchar *) g_hash_table_lookup ((GHashTable *) parameters, "password");
    server = (gchar *) g_hash_table_lookup ((GHashTable *) parameters, "server-address");
   
    g_set_error_val_if_fail (user != NULL || host != NULL || pass != NULL || server != NULL,
            FALSE, error, TAPIOCA_ERROR, TAPIOCA_ERROR_INVALID_PARAMETERS, "Invalid parameters");

    server_port = atoi (((gchar *) g_hash_table_lookup ((GHashTable *) parameters, "server-port")));
    use_encryption = atoi (((gchar *) g_hash_table_lookup ((GHashTable *) parameters, "use-encryption")));
    
    g_debug ("connecting to %s/%d using username %s at %s and encryption %s", 
            server, server_port, user, host, (use_encryption ? "on" : "off"));

    if (g_hash_table_lookup ((GHashTable *) parameters, "proxy-type")) {
        proxy_type = atoi ((gchar *) g_hash_table_lookup ((GHashTable *) parameters, "proxy-type"));
        if (proxy_type != TAPIOCA_PROXY_TYPE_NONE) {
            proxy = (gchar *) g_hash_table_lookup ((GHashTable *) parameters, "proxy-address");
            if (g_hash_table_lookup ((GHashTable *) parameters, "proxy-port"))
                proxy_port = atoi (((gchar *) g_hash_table_lookup ((GHashTable *) parameters, "proxy-port")));
            proxy_user = (gchar *) g_hash_table_lookup ((GHashTable *) parameters, "proxy-username");
            proxy_pass = (gchar *) g_hash_table_lookup ((GHashTable *) parameters, "proxy-password");
            
            g_debug ("proxy settings:\n\tserver/port: %s/%d\n\tusername: %s\ntype: %d", 
                    proxy, proxy_port, proxy_user, proxy_type);
        }
    }

    if (use_encryption) 
        xmpp_ssl_initialize ();

    /* start xmpp on a different thread */
    if (!thread) {
        thread = xmpp_thread_new ();
        xmpp_thread_manager_set_current (thread);
        xmpp_thread_start (thread);
    }
   
    xc->priv->client = xmpp_client_new ();
    xmpp_client_set_state_change_cb (xc->priv->client, state_change_cb, xc);
    xmpp_client_set_incoming_call_cb (xc->priv->client, incoming_call_cb, xc);
    xmpp_client_set_incoming_chat_cb (xc->priv->client, incoming_chat_cb, xc);

    xmpp_client_connect (xc->priv->client,
            user, host, resource, pass,
            server, server_port,
            proxy_type, proxy, proxy_port, proxy_user, proxy_pass,
            use_encryption);
  
    /* g_idle_add (loop_thread, NULL); */

    g_free (username);
    g_free (host);
    
    return TRUE;
}

static void
xmpp_connection_disconnect (TapiocaConnection *tc)
{
    XmppConnection *xc = XMPP_CONNECTION (tc);
    xmpp_client_disconnect (xc->priv->client);
}

static TapiocaSession *
xmpp_connection_create_session (TapiocaConnection *tc,
                                guint session_type,
                                GError **error)
{
    XmppConnection *xc = XMPP_CONNECTION (tc);
    TapiocaSession *ts;

    if (session_type == TAPIOCA_SESSION_TYPE_AUDIO) {
        ts = TAPIOCA_SESSION (xmpp_audio_session_new (xc->priv->client));
    }
    else if (session_type == TAPIOCA_SESSION_TYPE_TEXT) {
        ts = TAPIOCA_SESSION (xmpp_text_session_new (xc->priv->client));
    }
    else {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_SUPPORTED, 
                "Session of type %d is not supported by this plugin", session_type);
        return NULL;
    }
 
    if (!ts) {
        g_set_error (error, TAPIOCA_ERROR, TAPIOCA_ERROR_NOT_ENOUGH_MEMORY, 
                "Cannot create session instance. Memory low!");
        return NULL;
    }

    return ts;
}

static TapiocaPresence *
xmpp_connection_get_presence (TapiocaConnection *tc)
{
    XmppConnection *xc = XMPP_CONNECTION (tc);
    return TAPIOCA_PRESENCE (xc->priv->presence);
}

static TapiocaContactList *
xmpp_connection_get_contact_list (TapiocaConnection *tc)
{
    XmppConnection *xc = XMPP_CONNECTION (tc);
    return TAPIOCA_CONTACT_LIST (xc->priv->contact_list);
}

static TapiocaPrivacy *
xmpp_connection_get_privacy (TapiocaConnection *tc)
{
    XmppConnection *xc = XMPP_CONNECTION (tc);
    return TAPIOCA_PRIVACY (xc->priv->privacy);
}

static void
state_change_cb (xmpp_client_t client,
                 gpointer data, 
                 gint state,
                 gint error)
{
    TapiocaConnection *tc = TAPIOCA_CONNECTION (data);
    XmppConnection *xc = XMPP_CONNECTION (tc);

    switch (state) {
        case 1:
            g_debug ("connecting...");
            tapioca_connection_change_state (tc, TAPIOCA_CONNECTION_STATE_CONNECTING);
            break;
        case 2:
            g_debug ("logging in...");
            break;
        case 3:
            g_debug ("logged in...");
            xmpp_contact_list_initialize (xc->priv->contact_list, 
                    xmpp_client_get_roster (xc->priv->client));
            xmpp_privacy_initialize (xc->priv->privacy, 
                    xmpp_client_get_privacy (xc->priv->client));
            xmpp_presence_initialize (xc->priv->presence,
                    xmpp_client_get_presence_out (xc->priv->client),
                    xmpp_client_get_presence_push (xc->priv->client));
            xc->priv->keepalive_event_id = g_timeout_add (60000, (GSourceFunc) send_keepalive, xc->priv->client);
            tapioca_connection_change_state (tc, TAPIOCA_CONNECTION_STATE_CONNECTED);
            break;
        case 4: 
            {
                gint error_code = -1;
                
                g_debug ("logged out... %d", (int) error);
                
                if (xc->priv->keepalive_event_id != -1)
                    g_source_remove (xc->priv->keepalive_event_id);
                
                switch (error) {
                    case 0:  // buzz::XmppEngine::ERROR_NONE:
                        break;
                    case 1:  // buzz::XmppEngine::ERROR_XML:
                    case 2:  // buzz::XmppEngine::ERROR_STREAM:
                    case 3:  // buzz::XmppEngine::ERROR_VERSION:
                    case 5:  // buzz::XmppEngine::ERROR_TLS:
                    case 6:  // buzz::XmppEngine::ERROR_AUTH:
                    case 7:  // buzz::XmppEngine::ERROR_BIND:
                    case 8:  // buzz::XmppEngine::ERROR_CONNECTION_CLOSED:
                    case 9:  // buzz::XmppEngine::ERROR_DOCUMENT_CLOSED:
                    case 10: // buzz::XmppEngine::ERROR_SOCKET:
                        error_code = TAPIOCA_CONNECTION_ERROR_NETWORK_ERROR;
                        break;
                    case 4:  // buzz::XmppEngine::ERROR_UNAUTHORIZED:
                        error_code = TAPIOCA_CONNECTION_ERROR_AUTHENTICATION_FAILED;
                        break;
                    default:
                        error_code = TAPIOCA_ERROR_GENERIC;
                        break;
                }
                
                if (error_code != -1)
                    tapioca_connection_signal_error (tc, error_code);
                
                tapioca_connection_change_state (tc, TAPIOCA_CONNECTION_STATE_DISCONNECTED);
                break;
            }
    }
}

void
incoming_call_cb (xmpp_client_t client, 
                  gpointer data, 
                  const gchar *from,
                  xmpp_call_t call)
{
    XmppConnection *xc = XMPP_CONNECTION (data);
    TapiocaSession *ts = TAPIOCA_SESSION (
            xmpp_audio_session_new_with_call (client, call));

    g_debug ("incoming call from %s", from);

    tapioca_connection_signal_incoming_session (TAPIOCA_CONNECTION (xc), 
            from, ts);
}

void
incoming_chat_cb (xmpp_client_t client, 
                  gpointer data, 
                  const gchar *from,
                  xmpp_chat_t chat)
{
    XmppConnection *xc = XMPP_CONNECTION (data);
    TapiocaSession *ts = TAPIOCA_SESSION (
            xmpp_text_session_new_with_chat (client, chat));

    tapioca_connection_signal_incoming_session (TAPIOCA_CONNECTION (xc), 
            from, ts);
}

static gboolean
send_keepalive (xmpp_client_t client)
{
    xmpp_client_send_raw (client, "\t");
    return TRUE;
}

/*
static gboolean
loop_thread (gpointer data)
{
    xmpp_thread_loop (thread, 100);
    return TRUE;
}
*/
