/*
 * Copyright (c) 2005 INdT.
 * @author Luiz Augusto von Dentz <luiz.dentz@indt.org.br> 
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "xmpp_roster.h"

#include <talk/xmpp/constants.h>
#include <talk/xmpp/xmppclient.h>

namespace {
    
static const std::string NS_VCARD_TEMP("vcard-temp");

static const buzz::QName QN_VCARD(true, NS_VCARD_TEMP, "vCard");

}

XmppRoster::XmppRoster(buzz::Task *parent) 
    : buzz::XmppTask(parent, buzz::XmppEngine::HL_TYPE) 
{
}

XmppRoster::~XmppRoster()
{
}

void XmppRoster::AddContact(const buzz::Jid &jid, const std::string &name, 
    const std::vector<std::string> &groups)
{
    buzz::XmlElement *item = new buzz::XmlElement(buzz::QN_ROSTER_ITEM);
    buzz::XmlElement *subscribe = new buzz::XmlElement(buzz::QN_PRESENCE);
    std::vector<std::string>::const_iterator i;
    
    item->AddAttr(buzz::QN_JID, jid.Str());
    item->AddAttr(buzz::QN_NAME, name);
    for (i = groups.begin(); i != groups.end(); ++i) {
        buzz::XmlElement *item_group = new buzz::XmlElement(buzz::QN_ROSTER_GROUP);
        item_group->AddText(*i);
        if (item_group->BodyText() != buzz::STR_EMPTY)
            item->AddElement(item_group);
    }
    
    subscribe->AddAttr(buzz::QN_TO, jid.Str());
    subscribe->AddAttr(buzz::QN_TYPE, buzz::STR_SUBSCRIBE);

    SendStanza(subscribe);
    SendStanza(GenerateXml(buzz::STR_SET, item));
}

void XmppRoster::RemoveContact(const buzz::Jid &jid)
{
    buzz::XmlElement *item = new buzz::XmlElement(buzz::QN_ROSTER_ITEM);

    DenyContact(jid);
    
    item->AddAttr(buzz::QN_JID, jid.Str());
    item->AddAttr(buzz::QN_SUBSCRIPTION, buzz::STR_REMOVE);
        
    SendStanza(GenerateXml(buzz::STR_SET, item));
}

void XmppRoster::AuthorizeContact(const buzz::Jid &jid)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_PRESENCE);

    result->AddAttr(buzz::QN_TO, jid.Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_SUBSCRIBED);
    
    SendStanza(result);    
}

void XmppRoster::DenyContact(const buzz::Jid &jid)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_PRESENCE);

    result->AddAttr(buzz::QN_TO, jid.Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_UNSUBSCRIBED);
    
    SendStanza(result);
}

void XmppRoster::RequestContacts()
{       
    SendStanza(GenerateXml(buzz::STR_GET));
}

void XmppRoster::RequestInfo(const buzz::Jid &jid)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_IQ);
    
    result->AddAttr(buzz::QN_FROM, GetClient()->jid().Str());
    result->AddAttr(buzz::QN_TO, jid.BareJid().Str());
    result->AddAttr(buzz::QN_TYPE, buzz::STR_GET);
    result->AddAttr(buzz::QN_ID, GenerateId());

    result->AddElement(new buzz::XmlElement(QN_VCARD, false));
    result->AddAttr(buzz::QN_XMLNS, NS_VCARD_TEMP, 1);

    SendStanza(result);
}

bool XmppRoster::HandleStanza(const buzz::XmlElement *stanza) 
{
    /* Only handle Presence stanzas with types for subscriptions */
    if (stanza->Name() == buzz::QN_PRESENCE) {
        if (stanza->Attr(buzz::QN_TYPE) != buzz::STR_SUBSCRIBED && 
            stanza->Attr(buzz::QN_TYPE) != buzz::STR_SUBSCRIBE &&
            stanza->Attr(buzz::QN_TYPE) != buzz::STR_UNSUBSCRIBED)
            return false;
    }
    else if (stanza->Name() == buzz::QN_IQ) {
        if ((stanza->FirstElement() == NULL) ||
            (stanza->FirstElement()->Name() != buzz::QN_ROSTER_QUERY &&
    	     stanza->FirstElement()->Attr(buzz::QN_XMLNS) != buzz::Constants::ns_roster()) &&
             stanza->FirstElement()->Name() != QN_VCARD) { 
            return false;
        }

        if (stanza->FirstElement()->Name() == QN_VCARD &&
            stanza->Attr(buzz::QN_TYPE) != buzz::STR_RESULT) { 
            return false;
        }
    }
    else
        return false;
    QueueStanza(stanza);
    return true;
}

int XmppRoster::ProcessStart() 
{
    const buzz::XmlElement *stanza = NextStanza();
        
    if (stanza == NULL)
        return STATE_BLOCKED;

    if ((stanza->Name() == buzz::QN_IQ) && 
        (stanza->Attr(buzz::QN_TYPE) != buzz::STR_SET) &&
        (stanza->FirstNamed(buzz::QN_ROSTER_QUERY))) {
        crit_.Enter();
        const buzz::XmlElement *query = stanza->FirstNamed(buzz::QN_ROSTER_QUERY);
        /* Parse itens(contacts) */
        const buzz::XmlElement *items;
        for (items = query->FirstNamed(buzz::QN_ROSTER_ITEM); items;
            items = items->NextNamed(buzz::QN_ROSTER_ITEM)) {
            const std::string subscription = items->Attr(buzz::QN_SUBSCRIPTION);
            /* Parse itens(contacts) groups */
            if (items->FirstNamed(buzz::QN_ROSTER_GROUP)) {
                for (const buzz::XmlElement *groups = items->FirstNamed(buzz::QN_ROSTER_GROUP); groups;
                     groups = groups->NextNamed(buzz::QN_ROSTER_GROUP)) {
                    if (contact_cb_) {
                        contact_cb_((xmpp_roster_t) this, contact_data_,
                            items->Attr(buzz::QN_JID).c_str(),
                            items->Attr(buzz::QN_NAME).c_str(),
                            subscription.c_str(),
                            groups->BodyText().c_str());
                    }
                }
            }
            else {            
                if (contact_cb_) {
                    contact_cb_((xmpp_roster_t) this, contact_data_,
                        items->Attr(buzz::QN_JID).c_str(),
                        items->Attr(buzz::QN_NAME).c_str(),
                        subscription.c_str(),
                        "");
                }
            }
        }
        /* WORKAROUND: Send empty string to inform end of contact list */
        if (contact_cb_) {
            contact_cb_((xmpp_roster_t) this, contact_data_,
                "",
                "",
                "",
                "");
        }
        crit_.Leave();
    }
    else if (stanza->Name() == buzz::QN_PRESENCE) {
        if ((stanza->Attr(buzz::QN_TYPE) == buzz::STR_SUBSCRIBED) && authorize_cb_)
            authorize_cb_((xmpp_roster_t) this, authorize_data_,
                stanza->Attr(buzz::QN_FROM).c_str());
        else if ((stanza->Attr(buzz::QN_TYPE) == buzz::STR_UNSUBSCRIBED) && deny_cb_)
            deny_cb_((xmpp_roster_t) this, deny_data_, stanza->Attr(buzz::QN_FROM).c_str());
        else if ((stanza->Attr(buzz::QN_TYPE) == buzz::STR_SUBSCRIBE) && request_cb_)
            request_cb_((xmpp_roster_t) this, request_data_,
                stanza->Attr(buzz::QN_FROM).c_str());
    }
    else if ((stanza->Name() == buzz::QN_IQ) && 
             (stanza->Attr(buzz::QN_TYPE) == buzz::STR_RESULT) &&
             (stanza->FirstNamed(QN_VCARD))) {
        const buzz::XmlElement *vcard = stanza->FirstNamed(QN_VCARD);
        info_cb_ ((xmpp_roster_t) this, info_data_, 
                stanza->Attr(buzz::QN_FROM).c_str(), vcard->Str().c_str());
    }

    return STATE_START;
}

buzz::XmlElement 
*XmppRoster::GenerateXml(const std::string &type, buzz::XmlElement *item)
{
    buzz::XmlElement *result = new buzz::XmlElement(buzz::QN_IQ);
    buzz::XmlElement *query = new buzz::XmlElement(buzz::QN_ROSTER_QUERY);

    query->AddAttr(buzz::QN_XMLNS, buzz::Constants::ns_roster());
    if (item)
        query->AddElement(item);

    result->AddAttr(buzz::QN_FROM, GetClient()->jid().Str());
    result->AddAttr(buzz::QN_TYPE, type);
    result->AddAttr(buzz::QN_ID, GenerateId());
    result->AddElement(query);
    
    return result;
}
